from pyfyre.styles import Style
from pyfyre.states import State
from pyfyre.nodes.base import Node, Element
from typing import Any, Callable, Optional, Dict, List


class TextInput(Element):
	def __init__(
		self,
		children: Optional[Callable[[], List[Node]]] = None, *,
		placeholder: str = "",
		multiline: bool = False,
		styles: Optional[List[Style]] = None,
		states: Optional[List[State[Any]]] = None,
		attrs: Optional[Dict[str, str]] = None
	) -> None:
		attrs = attrs or {}
		attrs["placeholder"] = placeholder
		
		super().__init__(
			"textarea" if multiline else "input",
			children, styles=styles, states=states, attrs=attrs
		)
	
	@property
	def value(self) -> str:
		return str(self.dom.value)
	
	def set_value(self, value: Any) -> None:
		self.dom.value = str(value)
		self.update_dom()



import sys
from pyfyre.styles import Style
from pyfyre.states import State
from browser import DOMEvent, aio
from pyfyre.nodes import Node, Element
from typing import Optional, Dict, List, Callable, Any


class ListBuilder(Element):
	def __init__(
		self, children_builder: Callable[[int], Optional[Node]], *,
		count: Optional[int] = None,
		max_height: str = "300px",
		render_batch: int = 10,
		render_interval: float = 0,
		styles: Optional[List[Style]] = None,
		states: Optional[List[State[Any]]] = None,
		attrs: Optional[Dict[str, str]] = None
	) -> None:
		styles = styles or []
		self.children_builder = children_builder
		self.count = count
		self.render_batch = render_batch
		self.render_interval = render_interval
		
		self.index = 0
		self.rendered_children: List[Node] = []
		
		super().__init__(
			"div", lambda: self.rendered_children, attrs=attrs,
			styles=[Style(
				overflow_y="scroll",
				overflow_wrap="break-word",
				max_height=max_height
			)] + styles,
			states=states
		)
		
		def render_nodes(event: DOMEvent) -> None:
			el = event.target
			if el.scrollHeight - el.scrollTop - el.clientHeight < 1:
				rendered = self.render_next_children()
				if rendered:
					render_nodes(event)
		
		self.add_event_listener("scroll", render_nodes)
	
	def render_next_children(self) -> bool:
		prev_index = self.index
		
		for _ in range(self.render_batch):
			if self.count is not None:
				if self.index >= self.count:
					break
			
			try:
				child = self.children_builder(self.index)
			except BaseException:
				error_children = self.on_build_error(*sys.exc_info())
				child = Element("div", lambda: error_children)
			
			if child is None:
				break
			
			self.rendered_children.append(child)
			self.index += 1
		
		if self.index > prev_index:
			self.update_dom()
			return True
		
		return False
	
	def build_children(self, *, propagate: bool = True) -> None:
		async def async_wrapper() -> None:
			while self.dom.scrollHeight == self.dom.clientHeight:
				rendered = self.render_next_children()
				
				if not rendered:
					return
				
				await aio.sleep(self.render_interval)
		
		super().build_children(propagate=propagate)
		aio.run(async_wrapper())



from typing import Any


class EmptyObject:
	def __getattr__(self, attr: str) -> Any:
		return EmptyObject()
	
	def __call__(self, *args: Any, **kwargs: Any) -> Any:
		return EmptyObject()
	
	def __getitem__(self, key: str) -> Any:
		return EmptyObject()


Interpreter = EmptyObject()
Inspector = EmptyObject()



from enum import Enum
from typing import Callable
from browser import DOMEvent, window


class PyFyreEventType(Enum):
	pyfyreload = "pyfyreload"


def window_event_listener(
	event_type: str
) -> Callable[[Callable[[DOMEvent], None]], Callable[[DOMEvent], None]]:
	def window_event_listener_decorator(
		callback: Callable[[DOMEvent], None]
	) -> Callable[[DOMEvent], None]:
		window.bind(event_type, callback)
		return callback
	
	return window_event_listener_decorator



from pyfyre.styles import Style
from pyfyre.states import State
from abc import ABC, abstractmethod
from typing import Optional, Dict, List, Any
from pyfyre.nodes import Node, Element, FutureElement


class Widget(Element, ABC):
	def __init__(
		self, *,
		tag_name: str = "div",
		styles: Optional[List[Style]] = None,
		states: Optional[List[State[Any]]] = None,
		attrs: Optional[Dict[str, str]] = None
	) -> None:
		super().__init__(
			tag_name, self.build,
			styles=styles, states=states, attrs=attrs
		)
	
	@abstractmethod
	def build(self) -> List[Node]:
		raise NotImplementedError


class FutureWidget(FutureElement, ABC):
	def __init__(
		self, *,
		tag_name: str = "div",
		styles: Optional[List[Style]] = None,
		states: Optional[List[State[Any]]] = None,
		attrs: Optional[Dict[str, str]] = None
	) -> None:
		super().__init__(
			tag_name, self.build,
			styles=styles, states=states, attrs=attrs
		)
	
	@abstractmethod
	async def build(self) -> List[Node]:
		raise NotImplementedError



"""
	This script builds the app in the `_pyfyre` directory and starts a
	live server hosting that directory.
	The live server watches for changes in the `public` and `src` directories,
	as well as the `settings.py` and `template.html` files.
	When the live server detects changes, it will rebuild the app.
"""

from livereload import Server
from pyfyre_cli.build_app import build_app, bundle_scripts, create_pages


def run_app() -> None:
	build_app()
	server = Server()
	server.watch("public/", build_app)
	server.watch("src/", lambda: bundle_scripts(production=False))
	server.watch("settings.py", build_app)
	server.watch("template.html", lambda: create_pages(production=False))
	server.serve(root="_pyfyre")



from typing import List, Optional
from pyfyre.utils import EventMixin


class Style(EventMixin):
	@classmethod
	def from_styles(
		cls, styles: List["Style"], *, dynamic: bool = True
	) -> "Style":
		"""
			If [dynamic] is `True`, if one of the [styles] is updated,
			this [Style] will also update.
		"""
		
		s = Style()
		props = {}
		
		def update_style() -> None:
			s.props = Style.from_styles(styles, dynamic=False).props.copy()
			s.call_listeners()
		
		for style in styles:
			for prop, value in style.props.items():
				props[prop] = value
			
			if dynamic:
				style.add_listener(update_style)
		
		s.props = props
		return s
	
	def __init__(self, **props: str) -> None:
		super().__init__()
		self.props = props
	
	def __getitem__(self, prop: str) -> Optional[str]:
		return self.props.get(prop)
	
	def __setitem__(self, prop: str, value: str) -> None:
		self.props[prop] = value
		self.call_listeners()
	
	def __delitem__(self, prop: str) -> None:
		del self.props[prop]
		self.call_listeners()
	
	def css(self) -> str:
		return "; ".join([
			f"{prop.replace('_', '-')}: {value}"
			for prop, value in self.props.items()
		])



import os
import sys
import shutil
from typing import Iterator
from contextlib import contextmanager


@contextmanager
def in_path(path: str, *, append_to_sys_path: bool = False) -> Iterator[str]:
	original_path = os.getcwd()
	abspath = os.path.abspath(path)
	
	try:
		if append_to_sys_path:
			sys.path.append(abspath)
		
		os.chdir(abspath)
		yield abspath
	finally:
		if append_to_sys_path:
			sys.path.remove(abspath)
		
		os.chdir(original_path)


def empty_directory(path: str) -> None:
	for f in os.listdir(path):
		p = os.path.join(path, f)
		
		try:
			shutil.rmtree(p)
		except NotADirectoryError:
			os.remove(p)



from browser import aio
from pyfyre.styles import Style
from pyfyre.states import State
from pyfyre.nodes import Node, Element
from typing import Dict, List, Optional, Callable, Awaitable, Any
from pyfyre.exceptions import (
	FutureNoResult, FutureCancelled, FutureAlreadyDone
)


class FutureElement(Element):
	"""Error handling is not currently supported."""
	
	def __init__(
		self,
		tag_name: str,
		children: Callable[[], Awaitable[List[Node]]],
		*,
		styles: Optional[List[Style]] = None,
		states: Optional[List[State[Any]]] = None,
		attrs: Optional[Dict[str, str]] = None
	) -> None:
		self._is_done = False
		self._is_cancelled = False
		self._result: Optional[List[Node]] = None
		self._exception: Optional[BaseException] = None
		self._done_callbacks: List[Callable[["FutureElement"], None]] = []
		super().__init__(
			tag_name, lambda: self._process_result(),
			styles=styles, states=states, attrs=attrs
		)
		
		async def build_children(
			children: Callable[[], Awaitable[List[Node]]]
		) -> None:
			res = await children()
			
			# Set [ignore_done] to `True` to not have an error when this
			# future is already done before it sets the result.
			is_successful = self.set_result(res, ignore_done=True)
			
			if is_successful:
				self.update_dom()
		
		aio.run(build_children(children))
	
	def _process_result(self) -> List[Node]:
		try:
			return self.result()
		except FutureNoResult:
			return self.while_no_result()
	
	def while_no_result(self) -> List[Node]:
		from pyfyre.presets.loading import Loading
		return [Loading()]
	
	def result(self) -> List[Node]:
		if self.is_done():
			if self._result is not None:
				return self._result
			
			if self._exception is not None:
				raise self._exception
		
		if self.is_cancelled():
			raise FutureCancelled()
		
		raise FutureNoResult()
	
	def _mark_as_done(self) -> None:
		self._is_done = True
		
		for callback in self._done_callbacks:
			callback(self)
	
	def set_result(
		self, result: List[Node], *, ignore_done: bool = False
	) -> bool:
		if self.is_done():
			if ignore_done:
				return False
			
			raise FutureAlreadyDone()
		
		self._result = result
		self._mark_as_done()
		return True
	
	def set_exception(
		self, exception: BaseException, *, ignore_done: bool = False
	) -> bool:
		if self.is_done():
			if ignore_done:
				return False
			
			raise FutureAlreadyDone()
		
		self._exception = exception
		self._mark_as_done()
		return True
	
	def is_done(self) -> bool:
		return self.is_cancelled() or self._is_done
	
	def is_cancelled(self) -> bool:
		return self._is_cancelled
	
	def add_done_callback(
		self, callback: Callable[["FutureElement"], None]
	) -> None:
		self._done_callbacks.append(callback)
		
		if self.is_done():
			callback(self)
	
	def remove_done_callback(
		self, callback: Callable[["FutureElement"], None]
	) -> int:
		remaining_callbacks = []
		removed = 0
		
		for c in self._done_callbacks:
			if c == callback:
				removed += 1
			else:
				remaining_callbacks.append(c)
		
		self._done_callbacks = remaining_callbacks
		return removed
	
	def cancel(self) -> bool:
		if self.is_done():
			return False
		
		self._is_cancelled = True
		
		for callback in self._done_callbacks:
			callback(self)
		
		return True
	
	def exception(self) -> Optional[BaseException]:
		return self._exception



from pyfyre import render
from pyfyre.nodes import *


class App(Widget):
	def build(self) -> list[Node]:
		return [
			Element("h1", lambda: [Text("Welcome to PyFyre 🐍🔥")]),
			Element("p", lambda: [
				Text("Learn PyFyre by reading the "),
				Link(
					"https://pyfyre.netlify.app/",
					lambda: [Text("documentation")]
				),
				Text(". It is also advisable to learn "),
				Link("https://www.brython.info/", lambda: [Text("Brython")]),
				Text(" alongside PyFyre as it is built on top of Brython.")
			])
		]


render({"/": lambda: App()})



from typing import TypeVar, Generic
from pyfyre.utils import EventMixin

T = TypeVar("T")


class State(Generic[T], EventMixin):
	def __init__(self, value: T) -> None:
		super().__init__()
		self._value = value
	
	@property
	def value(self) -> T:
		return self._value
	
	def set_value(self, value: T) -> None:
		self._value = value
		self.call_listeners()



import os
import shutil
from pyfyre_cli.utils import in_path, empty_directory


def create_app(app_name: str, app_dir: str, *, dev_mode: bool = False) -> None:
	if not dev_mode:
		print(f"Creating your PyFyre project '{app_name}'...")
	
	project_path = os.path.join(app_dir, app_name)
	
	if os.path.isdir(project_path):
		prompt = "y" if dev_mode else input(
			f"Project '{app_name}' already exists.\nWant to overwrite "
			f"the directory? (y or n): "
		).lower()
		
		if prompt == "y":
			empty_directory(project_path)
		else:
			print("Aborting...")
			return
	
	with in_path(os.path.dirname(__file__)) as path:
		shutil.copytree(os.path.join(path, "user"), project_path, dirs_exist_ok=True)
	
	if not dev_mode:
		print("Project created successfully.")



from settings import ROUTES
from typing import Dict, Callable, Optional
from browser import document, window, DOMEvent
from pyfyre.events import window_event_listener
from pyfyre.nodes import Node, Element, TextNode


class RouteManager:
	_routes_builder: Dict[str, Callable[[], Node]] = {}
	_routes: Dict[str, Optional[Node]] = {}
	_root_node = document.select_one("#root")
	
	@staticmethod
	def initialize(routes: Dict[str, Callable[[], Node]]) -> None:
		RouteManager._routes_builder = routes
		
		@window_event_listener("popstate")
		def onpopstate(event: DOMEvent) -> None:
			RouteManager.change_route(window.location.pathname)
	
	@staticmethod
	def parse_route(route: str) -> str:
		el = document.createElement("a")
		el.href = route
		route = str(el.pathname)
		
		if route == "/":
			return route
		
		return str(el.pathname).rstrip("/")
	
	@staticmethod
	def get_node(route: str, *, parse_route: bool = True) -> Node:
		if parse_route:
			route = RouteManager.parse_route(route)
		
		node = RouteManager._routes.get(route)
		
		if node is None:
			route_builder = RouteManager._routes_builder.get(route)
			node = route_builder() if route_builder else None
			RouteManager._routes[route] = node
			
			if isinstance(node, Element):
				node.build_children()
		
		return node or Element("p", lambda: [TextNode("404: Page Not Found :(")])
	
	@staticmethod
	def render_route(route: str) -> None:
		node = RouteManager.get_node(route)
		RouteManager._root_node.clear()
		RouteManager._root_node.attach(node.dom)
	
	@staticmethod
	def change_route(route: str) -> None:
		route = RouteManager.parse_route(route)
		route_data = ROUTES.get(route) or {
			"title": "404: Page Not Found :("
		}
		
		document.title = route_data.get("title")
		RouteManager.render_route(route)



"""
Entry point for pyfyre installed via pip or setuptools.

Example:
```bash
pyfyre create hello-world
```
"""

import os
import sys
import pathlib
from typing import List, Optional

_HELP_MESSAGE = (
	"Manage your PyFyre projects.\n"
	"\n"
	"  Common Commands:\n"
	"    pyfyre help\n"
	"      Show this message.\n"
	"\n"
	"    pyfyre create [path]\n"
	"      Create a new PyFyre project in the specified path.\n"
	"      The directory of the path will be your project's name.\n"
	"\n"
	"    pyfyre run\n"
	"      Run the PyFyre project in the current directory in development mode.\n"
	"\n"
	"    pyfyre build\n"
	"      Build the PyFyre project in the current directory for production.\n"
)


def execute(args: Optional[List[str]] = None) -> None:
	args = args or sys.argv
	args_list: List[Optional[str]] = [None, None, None]
	
	for index, arg in enumerate(args):
		try:
			args_list[index] = arg
		except IndexError:
			break
	
	command = args_list[1] or "help"
	
	if command == "create":
		from pyfyre_cli.create_app import create_app
		app_name = args_list[2] or "pyfyre-app"
		app_path = pathlib.Path(os.path.abspath(app_name))
		create_app(
			app_path.parts[-1],
			os.path.join(*app_path.parts[:-1])
		)
	elif command == "run":
		from pyfyre_cli.run_app import run_app
		run_app()
	elif command == "build":
		from pyfyre_cli.build_app import build_app
		build_app(production=True)
	else:
		print(_HELP_MESSAGE)



"""
	This module contains the tools for building the app
	for development and production purposes.
	
	The build process includes but not limited to bundling of the files
	inside the `src` directory as a Brython package to make it usable for the web.
	
	All the build files are saved in the `build` directory so you can just
	serve or deploy the `build` directory to the web.
"""

import os
import sys
import shutil
import pathlib
import warnings
import importlib
import subprocess
from typing import List
from pyfyre_cli.utils import in_path
try:
	sys.path.append(os.getcwd())
	import settings
except ModuleNotFoundError:
	print("This directory is not a PyFyre project.")
	exit()


def _generate_page_head(*, production: bool) -> List[str]:
	head: List[str] = []
	
	if production:
		head.append('<script src="/modules.brython.js" defer="defer"></script>')
	else:
		head.append(
			'<script src="https://cdnjs.cloudflare.com/ajax/libs/brython/'
			'3.10.7/brython_stdlib.min.js" defer="defer"></script>'
		)
	
	if settings.PYTHON_DEPENDENCIES:
		head.append(
			'<script src="/cpython_packages.brython.js" defer="defer"></script>'
		)
	
	return head


def _generate_page_body(route_name: str) -> str:
	shutil.copytree("src", "__temp__", dirs_exist_ok=True)
	shutil.copy("settings.py", "__temp__")
	
	for module in ["browser", "interpreter", "javascript"]:
		shutil.copy(
			os.path.join(os.path.dirname(__file__), "copybin", f"{module}.py"),
			"__temp__"
		)
	
	with in_path("__temp__", append_to_sys_path=True):
		with warnings.catch_warnings():
			warnings.simplefilter("ignore")
			
			import pyfyre
			pyfyre.PRODUCTION = True
			
			import index
			from pyfyre.router import RouteManager
			
			# Set [parse_route] to `False` because the [get_node] method
			# uses a Brython function which will not work
			# if [parse_route] is `True` in this context because
			# we are running the app using Python and not Brython.
			# Brython functions like [createElement] will not work in this context
			# that's why we would not want to parse the route which uses
			# a Brython function (document.createElement).
			html = RouteManager.get_node(route_name, parse_route=False).html()
	
	shutil.rmtree("__temp__")
	return html


def create_pages(*, production: bool) -> None:
	importlib.reload(settings)
	
	with open("template.html") as file:
		html_template = file.read()
	
	for route, data in settings.ROUTES.items():
		directory = os.path.join(
			"build" if production else "_pyfyre",
			*route.split("/")
		)
		pathlib.Path(directory).mkdir(parents=True, exist_ok=True)
		
		with open(os.path.join(directory, "index.html"), "w") as file:
			head = _generate_page_head(production=production)
			html = html_template.format(
				prod_env=production,
				brython_options=str({
					"debug": 0 if production else 1,
					"cache": 1 if production else 0,
				}),
				title=data.get("title", "A PyFyre App"),
				icon=data.get("icon", ""),
				head="\n\t\t".join(head + data.get("head", [])),
				body=_generate_page_body(route) if production else ""
			)
			file.write(html)


def _cherry_pick_modules() -> None:
	shutil.copy(
		os.path.join(os.path.dirname(__file__), "copybin", "brython_stdlib.js"),
		"__temp__"
	)
	
	with in_path("__temp__"):
		# This command will detect the only Python modules that are used
		# in the project and will generate a cherry-picked standard libraries
		# for the project to run instead of using bloated `brython_stdlib.js`
		subprocess.run(["brython-cli", "make_modules"])
		os.rename("brython_modules.js", "modules.brython.js")
		shutil.copy("modules.brython.js", os.path.join("..", "build"))
		
		# Remove all files and folders except `index.py` to avoid duplicate VFS
		for file in os.listdir():
			if file == "index.py":
				continue
			
			if os.path.isfile(file):
				os.remove(file)
			else:
				shutil.rmtree(file)


def bundle_scripts(*, production: bool) -> None:
	shutil.copytree("src", "__temp__", dirs_exist_ok=True)
	shutil.copy("settings.py", "__temp__")
	
	temp_dir = os.path.abspath("__temp__")
	with in_path(os.path.join(os.path.dirname(__file__), "..")):
		shutil.copytree("pyfyre", os.path.join(temp_dir, "pyfyre"))
	
	if production:
		subprocess.run([
			"pyminify", "__temp__",
			"--in-place", "--remove-literal-statements"
		])
		
		subprocess.run([
			"autoflake", "__temp__", "-r", "--in-place", "--quiet",
			"--remove-unused-variables", "--remove-all-unused-imports",
			"--remove-duplicate-keys"
		])
		
		_cherry_pick_modules()
	
	with in_path("__temp__"):
		subprocess.run(["brython-cli", "make_package", "src"])
		shutil.copy(
			"src.brython.js",
			os.path.join("..", "build" if production else "_pyfyre")
		)
	
	shutil.rmtree("__temp__")


def add_cpython_packages(*, production: bool) -> None:
	importlib.reload(settings)
	
	for package_name in settings.PYTHON_DEPENDENCIES:
		subprocess.run(["brython-cli", "add_package", package_name])
	
	packages_dir = os.path.join("Lib", "site-packages")
	
	if not os.path.isdir(packages_dir):
		return
	
	with in_path(packages_dir):
		subprocess.run(["brython-cli", "make_package", "cpython_packages"])
		shutil.copy(
			"cpython_packages.brython.js",
			os.path.join("..", "..", "build" if production else "_pyfyre")
		)
	
	shutil.rmtree("Lib")


def build_app(*, production: bool = False) -> None:
	os.environ["PYTHONUTF8"] = "1"
	
	if production:
		print("Building app...")
	
	# Mirror the public files of the user to the build files
	shutil.copytree(
		"public",
		"build" if production else "_pyfyre",
		dirs_exist_ok=True
	)
	
	create_pages(production=production)
	bundle_scripts(production=production)
	add_cpython_packages(production=production)
	
	if production:
		print("App successfully built.")



from typing import Any, Type
from pyfyre.styles import Style
from pyfyre.nodes import Element, Text


class DebugError(Element):
	def __init__(
		self, *, exc_type: Type[BaseException],
		exc_value: BaseException, exc_traceback: str
	) -> None:
		super().__init__(
			"div",
			lambda: [
				Element("p", lambda: [Text("Unhandled Runtime Error")], styles=[Style(
					color="black", font_size="1.5rem", font_weight="bold",
					margin_top="0", margin_bottom="7px"
				)]),
				Element(
					"span", lambda: [Text(f"{exc_type.__name__}: {exc_value}")],
					styles=[Style(
						color="#ff3131", font_weight="bold",
						font_family="monospace", font_size="0.91rem"
					)]
				),
				Element("p", lambda: [Text("Traceback")], styles=[Style(
					color="black", font_size="1.3rem", font_weight="bold",
					margin_bottom="7px"
				)]),
				Element(
					"p",
					lambda: [Element("pre", lambda: [Text(exc_traceback)])],
					styles=[Style(
						background_color="black", color="white",
						padding="1px 15px", font_size="0.9rem", font_weight="normal"
					)]
				)
			],
			styles=[Style(
				padding="15px", background_color="white", box_shadow="0 0 10px #888888",
				border_top="5px solid #ff726f", border_radius="5px", font_family="Arial",
				display="inline-block"
			)]
		)


class ErrorMessage(Element):
	def __init__(self, message: Any) -> None:
		super().__init__(
			"div",
			lambda: [
				Element(
					"p",
					lambda: [Text(message)],
					styles=[Style(
						color="black", font_size="1rem", margin="0", font_weight="bold"
					)]
				)
			],
			styles=[Style(
				padding="15px", background_color="white", box_shadow="0 0 10px #888888",
				border_top="5px solid #ff726f", border_radius="5px", font_family="Arial",
				display="inline-block"
			)]
		)



from pyfyre.nodes import Node
from typing import Dict, Callable
from browser import window, DOMEvent
from pyfyre.router import RouteManager
from pyfyre.events import PyFyreEventType
from pyfyre.exceptions import PyFyreException
from .states import State
from .styles import Style

__all__ = [
	"PRODUCTION",
	"render",
	"State",
	"Style"
]

_rendered = False
PRODUCTION: bool = False


def _set_custom_window_events() -> Dict[str, DOMEvent]:
	events = {}
	
	for pyfyre_event in PyFyreEventType:
		event = window.Event.new("Event")
		event.initEvent(pyfyre_event.value, True, True)
		events[pyfyre_event.value] = event
	
	return events


def render(routes: Dict[str, Callable[[], Node]]) -> None:
	global _rendered
	if _rendered:
		raise PyFyreException("'render' function can only be called once")
	_rendered = True
	
	custom_events = _set_custom_window_events()
	
	RouteManager.initialize(routes)
	RouteManager.render_route(window.location.pathname)
	window.dispatchEvent(custom_events["pyfyreload"])



from browser import DOMEvent
from pyfyre.styles import Style
from pyfyre.states import State
from pyfyre.nodes.base import Node, Element
from typing import Callable, Optional, Dict, List, Any


class Button(Element):
	def __init__(
		self,
		onclick: Callable[[DOMEvent], None],
		children: Optional[Callable[[], List[Node]]] = None,
		*,
		styles: Optional[List[Style]] = None,
		states: Optional[List[State[Any]]] = None,
		attrs: Optional[Dict[str, str]] = None
	) -> None:
		super().__init__(
			"button", children, styles=styles, states=states, attrs=attrs
		)
		self.add_event_listener("click", onclick)



import sys
import pyfyre
import traceback
from typing import Type
from browser import document
from types import TracebackType
from pyfyre.styles import Style
from pyfyre.states import State
from abc import ABC, abstractmethod
from browser import DOMNode, DOMEvent
from typing import Any, Dict, List, Optional, Callable


class Node(ABC):
	def __init__(self) -> None:
		self.dom = self.create_dom()
	
	@abstractmethod
	def create_dom(self) -> DOMNode:
		raise NotImplementedError
	
	@abstractmethod
	def update_dom(self) -> None:
		raise NotImplementedError
	
	@abstractmethod
	def html(self) -> str:
		raise NotImplementedError


class Element(Node):
	def __init__(
		self,
		tag_name: str,
		children: Optional[Callable[[], List[Node]]] = None,
		*,
		styles: Optional[List[Style]] = None,
		states: Optional[List[State[Any]]] = None,
		attrs: Optional[Dict[str, str]] = None
	) -> None:
		self.tag_name = tag_name
		self.children: List[Node] = []
		self.style = Style.from_styles(styles) if styles else None
		self.states = states or []
		self.attrs = attrs or {}
		self._children_builder = children
		
		def update_style_attr() -> None:
			self.attrs = attrs or {}
			
			if self.style is not None:
				if "style" in self.attrs:
					self.attrs["style"] += f"; {self.style.css()}"
				else:
					self.attrs["style"] = self.style.css()
			
			if getattr(self, "dom", None) is not None:
				self.dom.setAttribute("style", self.attrs["style"])
		
		if self.style is not None:
			self.style.add_listener(update_style_attr)
			update_style_attr()
		
		for state in self.states:
			state.add_listener(self.update_dom)
		
		super().__init__()
	
	def _secure_build(self) -> List[Node]:
		if self._children_builder is None:
			return self.children
		
		try:
			return self._children_builder()
		except BaseException:
			return self.on_build_error(*sys.exc_info())
	
	def on_build_error(
		self, exc_type: Type[BaseException],
		exc_value: BaseException, exc_traceback: TracebackType
	) -> List[Node]:
		if pyfyre.PRODUCTION:
			from pyfyre.presets.errors import ErrorMessage
			return [ErrorMessage("An error occurred while loading this element")]
		
		from pyfyre.presets.errors import DebugError
		return [DebugError(
			exc_type=exc_type,
			exc_value=exc_value,
			exc_traceback=traceback.format_exc()
		)]
	
	def build_children(self, *, propagate: bool = True) -> None:
		self.children = self._secure_build()
		
		for child in self.children:
			self.dom.attach(child.dom)
			
			if propagate and isinstance(child, Element):
				child.build_children()
	
	def create_dom(self) -> DOMNode:
		el = document.createElement(self.tag_name)
		
		for attr_name, attr_value in self.attrs.items():
			el.setAttribute(attr_name, attr_value)
		
		return el
	
	def update_dom(self) -> None:
		self.dom.clear()
		self.update_attrs()
		self.build_children(propagate=False)
		
		for child in self.children:
			child.update_dom()
	
	def update_attrs(self) -> None:
		for attr_name, attr_value in self.attrs.items():
			self.dom.setAttribute(attr_name, attr_value)
	
	def html(self) -> str:
		def attributes() -> str:
			attrs = [
				f'{name}="{value}"'
				for name, value in self.attrs.items()
			]
			return " " + " ".join(attrs) if attrs else ""
		
		result = f"<{self.tag_name}{attributes()}>"
		
		for child in self.children:
			result += child.html()
		
		return result + f"</{self.tag_name}>"
	
	def add_event_listener(
		self, event_type: str, callback: Callable[[DOMEvent], None]
	) -> None:
		self.dom.bind(event_type, callback)


class TextNode(Node):
	def __init__(self, value: Any) -> None:
		if isinstance(value, State):
			self._value = str(value.value)
			value.add_listener(lambda: self.set_value(value.value))
		else:
			self._value = str(value)
		
		super().__init__()
	
	@property
	def value(self) -> str:
		return self._value
	
	def set_value(self, value: Any) -> None:
		self._value = str(value)
		self.update_dom()
	
	def create_dom(self) -> DOMNode:
		return document.createTextNode(self.value)
	
	def update_dom(self) -> None:
		self.dom.nodeValue = self.value
	
	def html(self) -> str:
		return self.value


E = Element
Text = TextNode



class PyFyreException(Exception):
	pass


class FutureElementException(PyFyreException):
	pass


class FutureNoResult(FutureElementException):
	pass


class FutureCancelled(FutureElementException):
	pass


class FutureAlreadyDone(FutureElementException):
	pass



from typing import Any


class EmptyObject:
	def __getattr__(self, attr: str) -> Any:
		return EmptyObject()
	
	def __call__(self, *args: Any, **kwargs: Any) -> Any:
		return EmptyObject()
	
	def __getitem__(self, key: str) -> Any:
		return EmptyObject()


import_js = EmptyObject()
py2js = EmptyObject()
this = EmptyObject()
Date = EmptyObject()
JSON = EmptyObject()
Math = EmptyObject()
NULL = EmptyObject()
Number = EmptyObject()
RegExp = EmptyObject()
String = EmptyObject()
UNDEFINED = EmptyObject()



import os
from typing import List
from setuptools import setup


def get_readme() -> str:
	with open("README.md") as file:
		return file.read()


def get_package_files(directory: str) -> List[str]:
	paths = []
	
	for (path, directories, filenames) in os.walk(directory):
		for filename in filenames:
			paths.append(os.path.join("..", path, filename))
	
	return paths


packages = ["pyfyre", "pyfyre_cli"]

setup(
	name="pyfyre",
	version="0.6.1",
	description="A fast, declarative, and incrementally adoptable Python web "
	"frontend framework for building reactive web user interfaces.",
	long_description=get_readme(),
	long_description_content_type="text/markdown",
	author="StackSpace",
	author_email="stackspace.ph@gmail.com",
	url="https://pyfyre.netlify.app/",
	download_url="https://pypi.org/project/pyfyre/",
	packages=packages,
	package_data={
		"": sum([get_package_files(directory) for directory in packages], [])
	},
	license="MIT",
	project_urls={
		"Source": "https://github.com/pyfyre/pyfyre",
		"Updates": "https://www.facebook.com/pyfyreframework/",
		"Documentation": "https://pyfyre.netlify.app/docs/"
	},
	entry_points={
		"console_scripts": ["pyfyre = pyfyre_cli:execute"]
	},
	install_requires=[i.strip() for i in open("requirements.txt").readlines()]
)



from pyfyre.styles import Style
from pyfyre.states import State
from pyfyre.nodes.base import Node, Element
from browser import DOMEvent, window, document
from typing import Optional, Dict, List, Callable, Any


class Link(Element):
	def __init__(
		self,
		href: str,
		children: Optional[Callable[[], List[Node]]] = None,
		*,
		styles: Optional[List[Style]] = None,
		states: Optional[List[State[Any]]] = None,
		attrs: Optional[Dict[str, str]] = None
	) -> None:
		self.href = href
		attrs = attrs or {}
		attrs["href"] = href
		super().__init__("a", children, styles=styles, states=states, attrs=attrs)
	
	@property
	def absolute_href(self) -> str:
		el = document.createElement("a")
		el.href = self.href
		return el.href
	
	def is_internal(self) -> bool:
		el = document.createElement("a")
		el.href = self.href
		return bool(el.host == window.location.host)


class RouterLink(Link):
	def __init__(
		self,
		href: str,
		children: Optional[Callable[[], List[Node]]] = None,
		*,
		styles: Optional[List[Style]] = None,
		attrs: Optional[Dict[str, str]] = None
	) -> None:
		super().__init__(href, children, styles=styles, attrs=attrs)
		
		def change_route(event: DOMEvent) -> None:
			# Import here due to cicular import problem
			from pyfyre.router import RouteManager
			
			window.history.pushState(None, None, self.absolute_href)
			event.preventDefault()
			RouteManager.change_route(href)
		
		self.add_event_listener("click", change_route)



from typing import Any


class EmptyObject:
	def __getattr__(self, attr: str) -> Any:
		return EmptyObject()
	
	def __call__(self, *args: Any, **kwargs: Any) -> Any:
		return EmptyObject()
	
	def __getitem__(self, key: str) -> Any:
		return EmptyObject()


alert = EmptyObject()
bind = EmptyObject()
confirm = EmptyObject()
console = EmptyObject()
document = EmptyObject()
DOMEvent = EmptyObject()
DOMNode = EmptyObject()
load = EmptyObject()
prompt = EmptyObject()
run_script = EmptyObject()
window = EmptyObject()
aio = EmptyObject()
ajax = EmptyObject()
html = EmptyObject()
local_storage = EmptyObject()
markdown = EmptyObject()
object_storage = EmptyObject()
session_storage = EmptyObject()
svg = EmptyObject()
template = EmptyObject()
timer = EmptyObject()
webcomponent = EmptyObject()
websocket = EmptyObject()
worker = EmptyObject()
widgets = EmptyObject()



from typing import List, Callable


class EventMixin:
	def __init__(self) -> None:
		self._listeners: List[Callable[[], None]] = []
	
	def call_listeners(self) -> None:
		for listener in self._listeners:
			listener()
	
	def add_listener(self, listener: Callable[[], None]) -> None:
		self._listeners.append(listener)
	
	def remove_listener(self, listener: Callable[[], None]) -> int:
		remaining_listeners = []
		removed = 0
		
		for c in self._listeners:
			if c == listener:
				removed += 1
			else:
				remaining_listeners.append(c)
		
		self._listeners = remaining_listeners
		return removed
